package com.tomtomwork.webfleet.components.jms.test;

import  java.io.File;
import  java.io.FileInputStream;
import  java.io.IOException;
import  java.io.InputStream;

import  java.text.DecimalFormat;

import  java.util.ArrayList;
import  java.util.Arrays;
import  java.util.Collections;
import  java.util.HashMap;
import  java.util.HashSet;
import  java.util.Iterator;
import  java.util.LinkedList;
import  java.util.List;
import  java.util.Map;
import  java.util.Properties;
import  java.util.Random;
import  java.util.Set;

import  javax.jms.Connection;
import  javax.jms.ConnectionFactory;
import  javax.jms.DeliveryMode;
import  javax.jms.Destination;
import  javax.jms.JMSException;
import  javax.jms.MapMessage;
import  javax.jms.MessageConsumer;
import  javax.jms.MessageProducer;
import  javax.jms.Session;

import  javax.naming.Context;
import  javax.naming.InitialContext;
import  javax.naming.NamingException;

import  org.apache.log4j.BasicConfigurator;
import  org.apache.log4j.ConsoleAppender;
import  org.apache.log4j.Level;
import  org.apache.log4j.Logger;
import  org.apache.log4j.PatternLayout;

/**
 * JMS load test class
 * @author rschuetz
 */

public class LoadTest
{
 // === Static ================================================================

 private static final Logger         LOGGER          = Logger.getLogger(LoadTest.class);
 private static final long           RECEIVE_TIMEOUT = 30000L;

 private static final String         ID_MESSAGEID    = "messageid",
                                     ID_SEQUENCEID   = "sequenceid",
                                     ID_TIMESTAMP    = "timestamp",
                                     ID_PAYLOAD      = "payload";

 // === Attributes ============================================================

 private final DecimalFormat         NBRFORMATTER = new DecimalFormat("###,###,###,###,###,###");

 private final Map<Integer, byte[]>  MESSAGES;

 private final Random                RANDOM;

 private final File                  STOPFILE;

 private final long                  STARTTIME;

 private final int                   MSGCOUNT,
                                     MSGCOUNTVARIANCE,
                                     STAGELOOPS,
                                     STAGES,
                                     MINMSGSIZE,
                                     MAXMSGSIZE,
                                     COMMITINTERVAL,
                                     RECONNECTINTERVAL;

 private final boolean               PERSISTANT_STORAGE;

 private       long                  totalNbrMessagesSent,
                                     totalNbrMessagesReceived,

                                     totalNbrMsgSequenceErrors,
                                     totalNbrMessagesDuplicate,
                                     totalNbrMessagesArrivedLate;

 private       Sender                sender;
 private       List<Receiver>        receivers;

 // === Constructor ===========================================================

 public LoadTest(String[] args)
 {
  int                   argIdx;
  byte[]                message;

  String                senderPropertiesFile, receiverPropertiesFiles[];

  Map <Integer, byte[]> messages;

  // Initialize logger

  BasicConfigurator.resetConfiguration();

  Logger.getRootLogger().addAppender(new ConsoleAppender(new PatternLayout("%-5p [%d{ISO8601}] %m%n")));
  Logger.getRootLogger().setLevel(Level.INFO);

  // Get parameters

  if(args.length == 12)
  {
   argIdx                     = 0;

   senderPropertiesFile       =                      args[argIdx ++];
   receiverPropertiesFiles    =                      args[argIdx ++].split(",");

   STOPFILE                   = new File            (args[argIdx ++]);
   MSGCOUNT                   = Integer.parseInt    (args[argIdx ++]);
   MSGCOUNTVARIANCE           = Integer.parseInt    (args[argIdx ++]);
   STAGES                     = Integer.parseInt    (args[argIdx ++]);
   STAGELOOPS                 = Integer.parseInt    (args[argIdx ++]);
   MINMSGSIZE                 = Integer.parseInt    (args[argIdx ++]);
   MAXMSGSIZE                 = Integer.parseInt    (args[argIdx ++]);
   COMMITINTERVAL             = Integer.parseInt    (args[argIdx ++]);
   RECONNECTINTERVAL          = Integer.parseInt    (args[argIdx ++]);
   PERSISTANT_STORAGE         = Boolean.parseBoolean(args[argIdx ++]);
  }
  else
  {
   throw new IllegalArgumentException("Usage: " + LoadTest.class.getSimpleName() +
                                      " <sender properties file> <receiver properties file(s); comma separated>" +
                                   " <stop file name> <max. message count in a queue (+/- variance)> <message count variance>" +
                                   " <# of stages> <# of loops adding/removing # of variance messages on queue per stage>" +
                                   " <minimum message size in bytes> <maximum message size in bytes>" +
                                   " <commit interval in # of messages processed> <reconnect interval in # of messages processed>" +
                                   " <persistant storage (true/false)>");
  }

  // Log parameters

  LOGGER.info("=== " + this.getClass().getSimpleName() + " ===");
  LOGGER.info("SENDERCONFIG               = " + senderPropertiesFile);
  LOGGER.info("RECEIVERCONFIGS            = " + Arrays.asList(receiverPropertiesFiles));
  LOGGER.info("STOPFILE                   = " + STOPFILE);
  LOGGER.info("MSGCOUNT                   = " + MSGCOUNT);
  LOGGER.info("MSGCOUNTVARIANCE           = " + MSGCOUNTVARIANCE);
  LOGGER.info("STAGES                     = " + STAGES);
  LOGGER.info("STAGELOOPS                 = " + STAGELOOPS);
  LOGGER.info("MINMSGSIZE                 = " + MINMSGSIZE);
  LOGGER.info("MAXMSGSIZE                 = " + MAXMSGSIZE);
  LOGGER.info("COMMITINTERVAL             = " + COMMITINTERVAL);
  LOGGER.info("RECONNECTINTERVAL          = " + RECONNECTINTERVAL);
  LOGGER.info("PERSISTANT_STORAGE         = " + PERSISTANT_STORAGE);

  // Initialize

  RANDOM                      = new Random();
  STARTTIME                   = System.currentTimeMillis();

  totalNbrMessagesSent        = 0L;
  totalNbrMessagesReceived    = 0L;

  totalNbrMsgSequenceErrors   = 0L;
  totalNbrMessagesDuplicate   = 0L;
  totalNbrMessagesArrivedLate = 0L;

  // Create sender

  sender = new Sender("S-", new File(senderPropertiesFile));

  // Create receivers

  receivers = new LinkedList<Receiver>();

  for(int receiverID = 0; receiverID < receiverPropertiesFiles.length; receiverID ++)
  {
   receivers.add(new Receiver("R" + receiverID, new File(receiverPropertiesFiles[receiverID])));
  }

  // Pre-create messages

  messages = new HashMap<Integer, byte[]>();

  for(int i = MINMSGSIZE; i <= MAXMSGSIZE; i ++)
  {
   message = new byte[i];
   RANDOM.nextBytes(message);
   messages.put(i, message);
  }

  MESSAGES = Collections.unmodifiableMap(messages);
 }

 // === Public ================================================================

 public static void main(String[] args)
 throws Exception
 {
  (new LoadTest(args)).runLoadTest();
 }

 void runLoadTest()
 {
  int  mainloop, stage;
  long stageLimit, nbrMessagesToSend, nbrMessagesToReceive;

  try
  {
   // Initial connect

                                      sender  .connectWithRetries();
   for(Receiver receiver : receivers) receiver.connectWithRetries();

   // Loop until specified runtime elapsed

   for(mainloop = 0; ; mainloop ++)
   {
    // Check for stop file

    if(STOPFILE.exists())
    {
     LOGGER.info("Stop file found; stopping ...");
     break;
    }

    // Fill queue in stages

    for(stage = 0; stage <= STAGES; stage ++)
    {
     // Fill queue to stage limit

     stageLimit        = (MSGCOUNT * stage) / STAGES;
     nbrMessagesToSend = stageLimit - (getCommonSenderMessageID() - getCommonReceiverMessageID());

     LOGGER.info("FILL: Filling queue: Loops: Main: " + mainloop + " Stage: " + stage + "/" + STAGES + "; limit " + NBRFORMATTER.format(stageLimit));

     if(nbrMessagesToSend > 0)
     {
      sender.sendMessages(nbrMessagesToSend);
     }
     else if(nbrMessagesToSend < 0)
     {
      // Queue is longer than expected; most likely caused by duplicate messages. Fail.

      throw new IllegalStateException("Queue too long (" + NBRFORMATTER.format(-nbrMessagesToSend)+ " messages too much); stopping.");
     }

     // Apply variance

     for(int stageLoop = 1; stageLoop <= STAGELOOPS; stageLoop ++)
     {
      if(stageLimit - MSGCOUNTVARIANCE < 0)
      {
       LOGGER.info("FILL/VAR: Loops: Main: " + mainloop + " Stage: " + stage + "/" + STAGES + " StageLoop: " + stageLoop + "/" + STAGELOOPS + "; sending.");
       sender.sendMessages(MSGCOUNTVARIANCE);

       LOGGER.info("FILL/VAR: Loops: Main: " + mainloop + " Stage: " + stage + "/" + STAGES + " StageLoop: " + stageLoop + "/" + STAGELOOPS + "; receiving.");
       for(Receiver receiver : receivers) receiver.receiveMessages(MSGCOUNTVARIANCE);
      }
      else
      {
       LOGGER.info("FILL/VAR: Loops: Main: " + mainloop + " Stage: " + stage + "/" + STAGES + " StageLoop: " + stageLoop + "/" + STAGELOOPS + "; sending [1].");
       sender.sendMessages(MSGCOUNTVARIANCE);

       LOGGER.info("FILL/VAR: Loops: Main: " + mainloop + " Stage: " + stage + "/" + STAGES + " StageLoop: " + stageLoop + "/" + STAGELOOPS + "; receiving.");
       for(Receiver receiver : receivers) receiver.receiveMessages(MSGCOUNTVARIANCE * 2);

       LOGGER.info("FILL/VAR: Loops: Main: " + mainloop + " Stage: " + stage + "/" + STAGES + " StageLoop: " + stageLoop + "/" + STAGELOOPS + "; sending [2].");
       sender.sendMessages(MSGCOUNTVARIANCE);
      }
     }
    }

    // Print totals

    LOGGER.info("FILL: Totals: # msgs sent: " + NBRFORMATTER.format(totalNbrMessagesSent) + " # msgs recv: " + NBRFORMATTER.format(totalNbrMessagesReceived));

    // Empty queue in stages

    for(stage = STAGES; stage >= 0; stage --)
    {
     // Empty queue to stage limit

     stageLimit           = (MSGCOUNT * stage) / STAGES;
     nbrMessagesToReceive = (getCommonSenderMessageID() - getCommonReceiverMessageID()) - stageLimit;

     LOGGER.info("EMPY: Emptying queue: Loops: Main: " + mainloop + " Stage: " + stage + "/" + STAGES + "; limit " + NBRFORMATTER.format(stageLimit));

     if(nbrMessagesToReceive > 0)
     {
      for(Receiver receiver : receivers) receiver.receiveMessages(nbrMessagesToReceive);
     }
     else if(nbrMessagesToReceive < 0)
     {
      // Queue is shorter than expected; most likely caused by lost messages. Fail

      throw new IllegalStateException("Queue too short(" + NBRFORMATTER.format(-nbrMessagesToReceive)+ " messages too little); stopping.");
     }

     // Apply variance

     for(int stageLoop = 1; stageLoop <= STAGELOOPS; stageLoop ++)
     {
      if(stageLimit - MSGCOUNTVARIANCE < 0)
      {
       LOGGER.info("EMPY/VAR: Loops: Main: " + mainloop + " Stage: " + stage + "/" + STAGES + " StageLoop: " + stageLoop + "/" + STAGELOOPS + "; sending.");
       sender.sendMessages(MSGCOUNTVARIANCE);

       LOGGER.info("EMPY/VAR: Loops: Main: " + mainloop + " Stage: " + stage + "/" + STAGES + " StageLoop: " + stageLoop + "/" + STAGELOOPS + "; receiving.");
       for(Receiver receiver : receivers) receiver.receiveMessages(MSGCOUNTVARIANCE);
      }
      else
      {
       LOGGER.info("EMPY/VAR: Loops: Main: " + mainloop + " Stage: " + stage + "/" + STAGES + " StageLoop: " + stageLoop + "/" + STAGELOOPS + "; sending [1].");
       sender.sendMessages(MSGCOUNTVARIANCE);

       LOGGER.info("EMPY/VAR: Loops: Main: " + mainloop + " Stage: " + stage + "/" + STAGES + " StageLoop: " + stageLoop + "/" + STAGELOOPS + "; receiving.");
       for(Receiver receiver : receivers) receiver.receiveMessages(MSGCOUNTVARIANCE * 2);

       LOGGER.info("EMPY/VAR: Loops: Main: " + mainloop + " Stage: " + stage + "/" + STAGES + " StageLoop: " + stageLoop + "/" + STAGELOOPS + "; sending [2].");
       sender.sendMessages(MSGCOUNTVARIANCE);
      }
     }
    }

    // Print totals

    LOGGER.info("EMPY: Totals: # msgs sent: " + NBRFORMATTER.format(totalNbrMessagesSent) + " # msgs recv: " + NBRFORMATTER.format(totalNbrMessagesReceived));
   }
  }
  catch(Exception e)
  {
   LOGGER.error("Exception caught", e);
  }
  finally
  {
   try
   {
                                       sender  .disconnect();
    for(Receiver receiver : receivers) receiver.disconnect();
   }
   catch(Exception e)
   {
    LOGGER.error("Exception caught while disconnecting", e);
   }
  }
  
  System.exit(0);
 }

 long getCommonSenderMessageID()
 {
  return(sender.msgSequenceID);
 }

 long getCommonReceiverMessageID()
 {
  Set<Long> msgSequenceIDs;

  // Collect message sequence IDs

  msgSequenceIDs = new HashSet<Long>();
  for(Receiver receiver : receivers) msgSequenceIDs.add(receiver.msgSequenceID);

  // Check

  if(msgSequenceIDs.size() != 1)
  {
   throw new IllegalStateException("Receivers have different msgSequenceIDs: " + msgSequenceIDs);
  }

  // Return ID

  return(msgSequenceIDs.iterator().next());
 }

 Properties loadProperties(File file)
 {
  Properties  properties;
  InputStream inputStream;

  // Initialize JNDI

  properties  = new Properties();
  inputStream = null;

  try
  {
   inputStream = new FileInputStream(file);
   properties.load(inputStream);
  }
  catch(IOException e)
  {
   LOGGER.warn("Unable to read file \"" + file.getAbsolutePath() + "\".", e);
  }
  finally
  {
   if(inputStream != null) try { inputStream.close(); } catch(Exception e) {}
  }

  return(properties);
 }

 // === Inner classes =========================================================

 private abstract class Connector
 {
        Connection          connection;
        Session             session;

  final String              name;
  final String              jndiFactoryInitial, jndiProviderURL, jndiConnectionFactoryAlias, jndiQueueAlias;
  final Map<String, String> addJndiICProperties;

        long                msgSequenceID;

  public Connector(final String name,
                   final String jndiFactoryInitial, final String jndiProviderURL, final String jndiConnectionFactoryAlias, final String jndiQueueAlias,
                   final Map<String, String> addJndiICProperties)
  {
   this.name                       = name;
   this.msgSequenceID              = 0;

   this.jndiFactoryInitial         = jndiFactoryInitial;
   this.jndiProviderURL            = jndiProviderURL;
   this.jndiConnectionFactoryAlias = jndiConnectionFactoryAlias;
   this.jndiQueueAlias             = jndiQueueAlias;
   this.addJndiICProperties        = Collections.unmodifiableMap(addJndiICProperties);

   // Log parameters

   LOGGER.info("=== " + this.getClass().getSimpleName() + " :: " + name + " ===");
   LOGGER.info("JNDIFACTORYINITIAL         = " + jndiFactoryInitial);
   LOGGER.info("JNDIPROVIDERURL            = " + jndiProviderURL);
   LOGGER.info("JNDICONNECTIONFACTORYALIAS = " + jndiConnectionFactoryAlias);
   LOGGER.info("JNDIQUEUEALIAS             = " + jndiQueueAlias);
   LOGGER.info("ADDJNDIICPROPERTIES        = " + addJndiICProperties);
  }

  public Connector(final String name, final File connectionProperties)
  {
   Properties          properties;
   Map<String, String> addJndiICProperties;

   // Set standard attributes

   this.name          = name;
   this.msgSequenceID = 0;

   // Load properties

   properties          = loadProperties(connectionProperties);
   addJndiICProperties = new HashMap<String, String>();

   // Set properties

   this.jndiFactoryInitial         = properties.getProperty("jndi.ic.factoryinitial",       null);
   this.jndiProviderURL            = properties.getProperty("jndi.ic.providerurl",          null);
   this.jndiConnectionFactoryAlias = properties.getProperty("jndi.alias.connectionfactory", null);
   this.jndiQueueAlias             = properties.getProperty("jndi.alias.queue",             null);

   for(String key : properties.stringPropertyNames())
   {
    if(key.startsWith("jndi.ic.add.")) addJndiICProperties.put(key.substring("jndi.ic.add.".length()), properties.getProperty(key));
   }

   this.addJndiICProperties = Collections.unmodifiableMap(addJndiICProperties);

   // Log parameters

   LOGGER.info("=== " + this.getClass().getSimpleName() + " :: " + name + " ===");
   LOGGER.info("JNDIFACTORYINITIAL         = " + jndiFactoryInitial);
   LOGGER.info("JNDIPROVIDERURL            = " + jndiProviderURL);
   LOGGER.info("JNDICONNECTIONFACTORYALIAS = " + jndiConnectionFactoryAlias);
   LOGGER.info("JNDIQUEUEALIAS             = " + jndiQueueAlias);
   LOGGER.info("ADDJNDIICPROPERTIES        = " + addJndiICProperties);
  }

  public abstract void connect()
  throws JMSException,
         NamingException;

  public void connectWithRetries()
  throws InterruptedException
  {
   for(;;)
   {
    try
    {
     connect(); return;
    }
    catch(Exception e)
    {
     LOGGER.error(name + ": Exception while connecting; retrying.", e);

     disconnect();
     Thread.sleep(1000);
    }
   }
  }

  public abstract void disconnect();

  public void reconnectWithRetries()
  throws InterruptedException
  {
   disconnect();
   connectWithRetries();
  }
 }

 private class Sender extends Connector
 {
  private MessageProducer producer;

  public Sender(final String name,
                final String jndiFactoryInitial, final String jndiProviderURL, final String jndiConnectionFactoryAlias, final String jndiQueueAlias,
                final Map<String, String> addJndiICProperties)
  {
   super(name, jndiFactoryInitial, jndiProviderURL, jndiConnectionFactoryAlias, jndiQueueAlias, addJndiICProperties);
  }

  public Sender(final String name, final File connectionProperties)
  {
   super(name, connectionProperties);
  }

  public void sendMessages(final long nbrMessages)
  throws InterruptedException
  {
   boolean             proceed, reconnect, commit;

   byte[]              payload;
   int                 nbrCommits, nbrReconnects;

   long                time, nbrMessagesSent, preBatchInboundMsgSequenceID, startQueueSize;
   MapMessage          mapMessage;

   Set<Long>           msgsInBatch;

   // Log

   LOGGER.info(name + ": Sending " + NBRFORMATTER.format(nbrMessages) + " messages.");

   // Process messages

   nbrMessagesSent              = 0L;
   time                         = System.currentTimeMillis();

   nbrCommits                   = 0;
   nbrReconnects                = 0;

   startQueueSize               = msgSequenceID - getCommonReceiverMessageID();

   preBatchInboundMsgSequenceID = msgSequenceID;
   msgsInBatch                  = new HashSet<Long>();

   while(nbrMessagesSent < nbrMessages)
   {
    // Select payload

    payload = MESSAGES.get(RANDOM.nextInt(MESSAGES.size()));

    // Create & send message

    proceed = false;

    do
    {
     try
     {
      mapMessage = session.createMapMessage();

      mapMessage.setStringProperty(ID_MESSAGEID,  this.getClass().getSimpleName() + "-" + STARTTIME + "#" + totalNbrMessagesSent);
      mapMessage.setLongProperty  (ID_TIMESTAMP,  System.currentTimeMillis());
      mapMessage.setLongProperty  (ID_SEQUENCEID, msgSequenceID);
      mapMessage.setBytes         (ID_PAYLOAD,    payload);

      producer.send(mapMessage);
      msgsInBatch.add(msgSequenceID);

      proceed = true;
     }
     catch(Exception e)
     {
      LOGGER.error(name + ": Exception while creating/sending message; reconnecting & retrying.", e);
      Thread.sleep(1000);

      // Reconnect

      try
      {
       reconnectWithRetries();
      }
      catch(Exception e1)
      {
       throw new IllegalStateException(name + ": reconnectWithRetries() failed after exception; stopping ...");
      }
     }
    } while(!proceed);

    // Adapt counters

    nbrMessagesSent      ++;
    totalNbrMessagesSent ++;
    msgSequenceID        ++;

    // Commit/Reconnect?

    reconnect = msgSequenceID % RECONNECTINTERVAL == 0;
    commit    = reconnect || msgSequenceID % COMMITINTERVAL == 0 || nbrMessagesSent == nbrMessages;

    // Commit

    if(commit)
    {
     if(COMMITINTERVAL >= 1000) LOGGER.info(name + ": Committing producer session; sequenceID: " + NBRFORMATTER.format(msgSequenceID) + " queueLength: " + NBRFORMATTER.format(msgSequenceID - getCommonReceiverMessageID()) + " uncommitted: " + NBRFORMATTER.format(msgsInBatch.size()));

     try
     {
      // Commit

      session.commit();

      // Mark messages as being in flight for each receiver

      for(Receiver receiver : receivers) receiver.msgsInFlight.addAll(msgsInBatch);
      msgsInBatch.clear();

      // Adapt pre-batch message sequence ID

      preBatchInboundMsgSequenceID = msgSequenceID;
     }
     catch(JMSException e)
     {
      LOGGER.error(name + ": JMSException while committing sent messages; resending messages.", e);

      // Commit failed, correct counters, messages must be resend

      nbrMessagesSent      -= msgsInBatch.size();
      totalNbrMessagesSent -= msgsInBatch.size();
      msgSequenceID         = preBatchInboundMsgSequenceID;

      msgsInBatch.clear();

      // Rollback

      try
      {
       session.rollback();
      }
      catch(Exception e1)
      {
       LOGGER.error(name + ": rollback() failed after commit() failure ...");
      }

      // Reconnect (test to see, if server becomes responsive again ...)

      try
      {
       reconnectWithRetries();
      }
      catch(Exception e1)
      {
       throw new IllegalStateException(name + ": reconnectWithRetries() failed after rollback(); stopping ...");
      }
     }

     nbrCommits ++;
    }

    if(reconnect)
    {
     // Disconnect/Reconnect

     proceed = false;

     do
     {
      try
      {
       reconnectWithRetries();
       proceed = true;
      }
      catch(Exception e)
      {
       LOGGER.error(name + ": Exception while reconnecting; retrying.", e);
       Thread.sleep(1000);
      }
     } while(!proceed);

     nbrReconnects ++;
    }
   }

   time = System.currentTimeMillis() - time;

   // Log

   LOGGER.info(name + ": SEND done. #msgs: " + NBRFORMATTER.format(nbrMessagesSent) + " inflight: " + NBRFORMATTER.format(startQueueSize) + " - " + NBRFORMATTER.format(msgSequenceID - getCommonReceiverMessageID()) + " speed: " + NBRFORMATTER.format((1000L * nbrMessagesSent) / time) + " msgs/s # commits: " + NBRFORMATTER.format(nbrCommits) + " #reconnects: " + NBRFORMATTER.format(nbrReconnects));
   LOGGER.info(name + ": Totals: #msgs sent: " + NBRFORMATTER.format(totalNbrMessagesSent));
  }

  public void connect()
  throws JMSException,
         NamingException
  {
   Properties        jndiProperties;
   InitialContext    jndiContext;
   ConnectionFactory connectionFactory;
   Destination       destination;

   // Log

   LOGGER.info(name + ": Connecting.");

   // Initialize JNDI

   jndiProperties = new Properties();

   // Add standard properties

   jndiProperties.setProperty(Context.INITIAL_CONTEXT_FACTORY, jndiFactoryInitial);
   jndiProperties.setProperty(Context.PROVIDER_URL,            jndiProviderURL);

   // Add additional properties

   for(Map.Entry<String, String> entry : addJndiICProperties.entrySet())
   {
    jndiProperties.setProperty(entry.getKey(), entry.getValue());
   }

   jndiContext = new InitialContext(jndiProperties);

   // Lookup connection factory and queue

   connectionFactory = ((ConnectionFactory)(jndiContext.lookup(jndiConnectionFactoryAlias)));
   destination       = ((Destination      )(jndiContext.lookup(jndiQueueAlias)));

   // Create producer connection, session and message producer

   connection = connectionFactory.createConnection();
   session    = connection.createSession(true, 0);

   producer = session.createProducer(destination);
   producer.setDeliveryMode(PERSISTANT_STORAGE ? DeliveryMode.PERSISTENT : DeliveryMode.NON_PERSISTENT);

   connection.start();

   // Log

   LOGGER.info(name + ": Connected.");
  }

  public void disconnect()
  {
   // Log

   LOGGER.info(name + ": Disconnecting.");

   // Close assets

   if(producer   != null) { try { producer  .close(); } catch(JMSException e) { LOGGER.error("Error while disconnecting producer",   e); } producer   = null; }
   if(session    != null) { try { session   .close(); } catch(JMSException e) { LOGGER.error("Error while disconnecting session",    e); } session    = null; }
   if(connection != null) { try { connection.close(); } catch(JMSException e) { LOGGER.error("Error while disconnecting connection", e); } connection = null; }

   // Log

   LOGGER.info(name + ": Disconnected.");
  }
 }

 private class Receiver extends Connector
 {
  MessageConsumer consumer;

  Set<Long>       msgsInFlight;
  Set<Long>       msgsLost;

  public Receiver(final String name,
                  final String jndiFactoryInitial, final String jndiProviderURL, final String jndiConnectionFactoryAlias, final String jndiQueueAlias,
                  final Map<String, String> addJndiICProperties)
  {
   super(name, jndiFactoryInitial, jndiProviderURL, jndiConnectionFactoryAlias, jndiQueueAlias, addJndiICProperties);

   this.msgsInFlight  = new HashSet<Long>(MSGCOUNT);
   this.msgsLost      = new HashSet<Long>();
  }

  public Receiver(final String name, final File connectionProperties)
  {
   super(name, connectionProperties);

   this.msgsInFlight  = new HashSet<Long>(MSGCOUNT);
   this.msgsLost      = new HashSet<Long>();
  }

  public void receiveMessages(final long nbrMessages)
  throws InterruptedException
  {
   boolean             proceed, reconnect, commit;

   long                nbrMessagesReceived, preBatchOutboundMsgSequenceID, receivedMsgSequenceID, time, startQueueSize;
   int                 nbrCommits, nbrReconnects;

   MapMessage          mapMessage;

   List<Long>          msgsInBatch;

   // Log

   LOGGER.info(name + ": Receiving " + NBRFORMATTER.format(nbrMessages) + " messages, current queue size: " + NBRFORMATTER.format(getCommonSenderMessageID() - msgSequenceID));

   // Process messages

   nbrMessagesReceived           = 0L;
   time                          = System.currentTimeMillis();

   nbrCommits                    = 0;
   nbrReconnects                 = 0;

   startQueueSize                = getCommonSenderMessageID() - msgSequenceID;

   preBatchOutboundMsgSequenceID = msgSequenceID;
   msgsInBatch                   = new ArrayList<Long>(COMMITINTERVAL);

   while(nbrMessagesReceived < nbrMessages)
   {
    // Receive

    proceed               = false;

    mapMessage            = null;
    receivedMsgSequenceID = 0L;

    do
    {
     try
     {
      if((mapMessage = ((MapMessage)(consumer.receive(RECEIVE_TIMEOUT)))) != null)
      {
       receivedMsgSequenceID = mapMessage.getLongProperty(ID_SEQUENCEID);
       proceed               = true;
      }
      else
      {
       // No message after timeout; reconnect might help

       LOGGER.error(name + ": messageConsumer#receive() did not return any new messages. Reconnecting.");

       try
       {
        reconnectWithRetries();
       }
       catch(Exception e1)
       {
        throw new IllegalStateException(name + ": reconnectWithRetries() failed after receive() did not return any new messages; stopping ...");
       }
      }
     }
     catch(Exception e)
     {
      LOGGER.error(name + ": Exception while receiving message; reconnecting & retrying.", e);
      Thread.sleep(1000);

      // Reconnect

      try
      {
       reconnectWithRetries();
      }
      catch(Exception e1)
      {
       throw new IllegalStateException(name + ": reconnectWithRetries() failed after exception; stopping ...");
      }
     }
    } while(!proceed);

    // Message received

    if(mapMessage != null)
    {
     // Store message ID

     msgsInBatch.add(msgSequenceID);

     // Compare sequence IDs

     if(receivedMsgSequenceID != msgSequenceID)
     {
      // Message sequence broken. Log.

      LOGGER.error(name + ": Received message != expected message. recv: " + NBRFORMATTER.format(receivedMsgSequenceID) + " exp: " + NBRFORMATTER.format(msgSequenceID));

      // Re-sync

      msgSequenceID = receivedMsgSequenceID;

      // Count sequence errors

      totalNbrMsgSequenceErrors ++;
     }

     // Update counters

     nbrMessagesReceived      ++;
     totalNbrMessagesReceived ++;
     msgSequenceID            ++;
    }

    // Commit/Reconnect?

    reconnect = msgSequenceID % RECONNECTINTERVAL == 0;
    commit    = reconnect || msgSequenceID % COMMITINTERVAL == 0 || nbrMessagesReceived == nbrMessages;

    // Commit

    if(commit)
    {
     // Commit

     if(COMMITINTERVAL >= 1000) LOGGER.info(name + ": Committing consumer session; sequenceID: " + NBRFORMATTER.format(msgSequenceID) + " queueLength: " + NBRFORMATTER.format(getCommonSenderMessageID() - msgSequenceID) + " uncommitted: " + NBRFORMATTER.format(msgsInBatch.size()));

     try
     {
      // Commit

      session.commit();

      // Check for duplicate messages, remove processed messages from global set

      for(final Long currMsgSequenceID : msgsInBatch)
      {
       if(!msgsInFlight.remove(currMsgSequenceID))
       {
        // Late arrival?

        if(msgsLost.remove(currMsgSequenceID))
        {
         LOGGER.error(name + ": Message arrived late: " + NBRFORMATTER.format(currMsgSequenceID));
         totalNbrMessagesArrivedLate ++;
        }
        else
        {
         LOGGER.error(name + ": Duplicate message: " + NBRFORMATTER.format(currMsgSequenceID));
         totalNbrMessagesDuplicate ++;
        }
       }
      }

      msgsInBatch.clear();

      // Adapt pre-batch message sequence ID

      preBatchOutboundMsgSequenceID = msgSequenceID;
     }
     catch(JMSException e)
     {
      LOGGER.error(name + ": JMSException while committing received messages; re-receiving messages.", e);

      // Commit failed, correct counters, ALL messages should be re-received.

      nbrMessagesReceived      -= msgsInBatch.size();
      totalNbrMessagesReceived -= msgsInBatch.size();
      msgSequenceID             = preBatchOutboundMsgSequenceID;

      msgsInBatch.clear();

      // Rollback

      try
      {
       session.rollback();
      }
      catch(Exception e1)
      {
       LOGGER.error(name + ": rollback() failed after commit() failure ...");
      }
      
      // Reconnect (test to see, if server becomes responsive again ...)

      try
      {
       reconnectWithRetries();
      }
      catch(Exception e1)
      {
       throw new IllegalStateException(name + ": reconnectWithRetries() failed after rollback(); stopping ...");
      }
     }

     nbrCommits ++;
    }

    if(reconnect)
    {
     // Disconnect/Reconnect

     try
     {
      reconnectWithRetries();
     }
     catch(Exception e)
     {
      throw new IllegalStateException(name + ": Scheduled reconnectWithRetries() failed; stopping ...");
     }

     nbrReconnects ++;
    }
   }

   time = System.currentTimeMillis() - time;

   // Check for lost messages. We consider messages as being lost, if they did not appear on the outbound side
   // after sending 2 x max. queue length messages

   for(Iterator<Long> msgsInFlightIterator = msgsInFlight.iterator(); msgsInFlightIterator.hasNext(); )
   {
    // Get ID

    final Long currMsgSequenceID = msgsInFlightIterator.next();

    // Lost?

    if(currMsgSequenceID < msgSequenceID - 2 * MSGCOUNT)
    {
     // Log

     LOGGER.error("Lost message: " + NBRFORMATTER.format(currMsgSequenceID));

     // Move to lost messages

     msgsLost.add(currMsgSequenceID);
     msgsInFlightIterator.remove();
    }
   }

   // Log

   LOGGER.info(name + ": RECV done. #msgs: " + NBRFORMATTER.format(nbrMessagesReceived) + " inflight: " + NBRFORMATTER.format(startQueueSize) + " - " + NBRFORMATTER.format(getCommonSenderMessageID() - msgSequenceID) + " speed: " + NBRFORMATTER.format((1000L * nbrMessagesReceived) / time) + " msgs/s # commits: " + NBRFORMATTER.format(nbrCommits) + " #reconnects: " + NBRFORMATTER.format(nbrReconnects));
   LOGGER.info(name + ": Totals: #msgs recv: " + NBRFORMATTER.format(totalNbrMessagesReceived) + " inflight: " + NBRFORMATTER.format(msgsInFlight.size()) + " lost: " + NBRFORMATTER.format(msgsLost.size()) + " dupl: " + NBRFORMATTER.format(totalNbrMessagesDuplicate) + " late: " + NBRFORMATTER.format(totalNbrMessagesArrivedLate) + " #seq errors: " + NBRFORMATTER.format(totalNbrMsgSequenceErrors));
  }

  public void connect()
  throws JMSException,
         NamingException
  {
   Properties        jndiProperties;
   InitialContext    jndiContext;
   ConnectionFactory connectionFactory;
   Destination       destination;

   // Log

   LOGGER.info(name + ": Connecting.");

   // Initialize JNDI

   jndiProperties = new Properties();

   // Add standard properties

   jndiProperties.setProperty(Context.INITIAL_CONTEXT_FACTORY, jndiFactoryInitial);
   jndiProperties.setProperty(Context.PROVIDER_URL,            jndiProviderURL);

   // Add additional properties

   for(Map.Entry<String, String> entry : addJndiICProperties.entrySet())
   {
    jndiProperties.setProperty(entry.getKey(), entry.getValue());
   }

   jndiContext = new InitialContext(jndiProperties);

   // Lookup connection factory and queue

   connectionFactory = ((ConnectionFactory)(jndiContext.lookup(jndiConnectionFactoryAlias)));
   destination       = ((Destination      )(jndiContext.lookup(jndiQueueAlias)));

   // Create producer connection, session and message producer

   connection = connectionFactory.createConnection();
   session    = connection.createSession(true, 0);

   consumer = session.createConsumer(destination);

   connection.start();

   // Log

   LOGGER.info(name + ": Connected.");
  }

  public void disconnect()
  {
   // Log

   LOGGER.info(name + ": Disconnecting.");

   // Close assets

   if(consumer   != null) { try { consumer  .close(); } catch(JMSException e) { LOGGER.error("Error while disconnecting consumer",   e); } consumer   = null; }
   if(session    != null) { try { session   .close(); } catch(JMSException e) { LOGGER.error("Error while disconnecting session",    e); } session    = null; }
   if(connection != null) { try { connection.close(); } catch(JMSException e) { LOGGER.error("Error while disconnecting connection", e); } connection = null; }

   // Log

   LOGGER.info(name + ": Disconnected.");
  }
 }
}
